/*
 * frag_find: given a file, find all of its fragments on the hard drive.
 * Automatically ignore popular fragments. 
 *
 * Performance Log:
 * Version 1.1.1 - 20724 SHA1 calcuations
 */

#include "bloom.h"
#include "bloomconfig.h"
#include "aftimer.h"

#define __STDC_FORMAT_MACROS

#include <assert.h>
#include <getopt.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <signal.h>
#include <string.h>
#include <cstring>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#if defined(HAVE_SYS_DISK_H) 
#include <sys/disk.h>
#endif
#if defined(HAVE_SYS_MOUNT_H)
#include <sys/mount.h>
#endif

#include <map>
#include <vector>

#define USE_BITSET
#ifdef USE_BITSET
#include <bitset>
#endif

#ifndef DEFAULT_BLOCKSIZE
#define DEFAULT_BLOCKSIZE 512
#endif

#include "blockfile.h"

#ifdef HAVE_ERR_H
#include <err.h>
#else
void err(int eval,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,": %s\n",strerror(errno));
  exit(eval);
}
#endif

#ifndef HAVE_ATOI64
int64_t atoi64(const char *buf)
{
    int64_t ret=0;
    sscanf(buf,"%"PRId64,&ret);
    return ret;
}
#endif


u_int blocksize = DEFAULT_BLOCKSIZE;

uint64_t blocknumber=0;
void status(int n)
{
    printf("current block: %"PRId64"\n",blocknumber);
}

/* SHA1 class.
 * This class both represents the SHA1 and can calculate it from a supplied buffer.
 *
 * This instrumented sha1 keeps track of how many times it has been called and 
 * has a bypass for calculating SHA1 of all NULLs.
 */
uint64_t sha1_ctr=0;			// number of SHA1 operations performed
static u_char sha1_512nulls[20];
class sha1_t {
public:
    static bool iszero(const u_char *buf,size_t bufsize){
	for(u_int i=0;i<bufsize;i++){
	    if(buf[i]!=0) return false;
	}
	return true;
    }
    u_char sha1[20];			// the sha1 value
    char hexbuf[41];
    SHA_CTX ctx;
    sha1_t(){
	SHA_Init(&ctx);
	memset(sha1,0,sizeof(sha1));
    }
    void update(const u_char *buf,size_t bufsize){
	SHA_Update(&ctx,buf,bufsize);
	printf("update; bufsize=%d\n",(int)bufsize);
    }
    void final(){
	SHA_Final(sha1,&ctx);
    }

    sha1_t(const u_char *buf,size_t bufsize){

	/* First time through find the SHA1 of 512 NULLs */
	if(sha1_ctr==0){
	    u_char b2[512];
	    memset(b2,0,sizeof(b2));
	    SHA1(b2,sizeof(b2),sha1_512nulls);
	    sha1_ctr++;
	}

	/* If the input is 512 bytes long and all NULLs, use our SHA1 of 512 NULLs */
	if(bufsize==512 && iszero(buf,bufsize)){
	    memcpy(sha1,sha1_512nulls,sizeof(sha1));
	    return;
	}

	SHA1((u_char *)buf,bufsize,sha1);
	sha1_ctr++;
    }

    /** hexbuf returns the hex representation of this string.
     * Not that this is not threadsafe.
     */     
    const char *hex() {
	for(int i=0;i<20;i++){
	    printf("%d=%x\n",i,sha1[i]);
	    sprintf(hexbuf+i*2,"%02x",sha1[i]);
	}
	return hexbuf;
    }
    bool operator<(const sha1_t &s2) const
    {
	return memcmp(this->sha1,s2.sha1, 20) < 0;
    }
};

/** blocklist is a list of block numbers.
 */

class blocklist_t:public std::vector<int64_t> {
public:
    /** return true if the blocklist contains a given block number */
    bool contains(int64_t block) const {
	for(blocklist_t::const_iterator i = this->begin();
	    i != this->end();
	    i++){
	    if(*i == block) return true;
	}
	return false;
    }
};

/* A filemap is an array of arrays. The key is the block number of the mapped file.
 * the value is an array of all the locations where the block appars
 */
class filemap_t:public std::vector<blocklist_t> {
public:
    /* Clean the filemap. If any sector in the input file is in more than
     * one place in the image, check to see if the previous next is in just one place.
     * If so, then only pick the sector that aligns.
     */
    void clean(){
	for(int64_t b=1;b<(int64_t)this->size();b++){
	    if((*this)[b-1].size()==1 && (*this)[b].size()>1){
		int64_t predicted = (*this)[b-1].front()+1;
		if((*this)[b].contains(predicted)){
		    (*this)[b].clear();
		    (*this)[b].push_back(predicted);
		}
	    }
	}
	
	/* Clean in the reverse direction */
	for(int64_t b=this->size()-1;b>=0;b--){
	    if((*this)[b+1].size()==1 && (*this)[b].size()>1){
		int64_t predicted = (*this)[b+1].front()-1;
		if((*this)[b].contains(predicted)){
		    (*this)[b].clear();
		    (*this)[b].push_back(predicted);
		}
	    }
	}
    }

    /* Return the length of a given mapping run --- the number of sequential blocks
     * starting at a given location.
     */
    int64_t runlen(int64_t target_block,int64_t image_block){
	int64_t len = 1;
	for(len = 1;target_block + len < (int64_t)this->size(); len++){
	    if(! (*this)[target_block+len].contains(image_block+len)){
		return len;
	    }
	}
	return len;
    }

    int64_t found_count(){
	int64_t count = 0;
	for(int64_t b=0;b<(int64_t)this->size();b++){
	    if ((*this)[b].size()>0) count++;
	}
	return count;
    }

    double found_percent(){
	return found_count() * 100.0 / this->size();
    }

    void print(){
	printf("Target Block(s)     Found at image block\n");
	for(int64_t tb=0;tb<(int64_t)this->size();){
	    if ((*this)[tb].size()==0){
		tb++;			// target block not in image
		continue;
	    }
	    int64_t first_ib = (*this)[tb].front();
	    int64_t runlen = this->runlen(tb,first_ib);
	    if(runlen>1){
		printf("%10"PRId64"-%-10"PRId64"    %10"PRId64"-%-10"PRId64"   (%"PRId64" blocks)\n",
		       tb,tb+runlen-1,
		       first_ib,
		       first_ib+runlen-1,
		       runlen);
		tb += runlen;
		continue;
	    }
	    bool first=true;
	    for(blocklist_t::const_iterator i = (*this)[tb].begin();
		i != (*this)[tb].end();
		i++){
		if(first){
		    printf("%10"PRId64"               ",tb);
		    first = false;
		}
		printf(" %10"PRId64" ",*i);
	    }
	    if(first==false) printf("\n");
	    tb++;
	}
    }
};



/** sha1map maps individual sha1's to a list of blocks
 */

class sha1map_t:public std::map<class sha1_t,blocklist_t> {
};

/** The prefilter */

class prefilter_t:public std::bitset<16777216> {
    /** return a very fast-to-calculate 24-bit value from a 512-byte buffer*/
    static uint32_t val3char(const u_char *buf) {
	uint32_t counter = 0;
	for(int i=0;i<512;i++){
	    counter = ((counter << 1) | (counter>>31)); // rotate and Xor-in a bit
	    counter = counter ^ (counter >> 9 & 0x1) ^ buf[i];
	}
	return counter & 0xffffff;
    }
public:
    int64_t hits;
    prefilter_t():hits(0){
    }
    double utilization() {
	int64_t seen = 0;
	for(int i=0;i<16777216;i++){
	    if((*this)[i]) seen++;
	}
	return (double)seen / 16777216.0;
    }
    void setbuf(const u_char *buf){
	(*this)[val3char(buf)] = 1;
    }
    bool querybuf(const u_char *buf){
	bool ret = (*this)[val3char(buf)] == 1;
	if(ret) hits++;
	return ret;
    }
};


int opt_M = 32;
void usage()
{
    printf("usage: frag_find [options] <raw disk file> <target1> [<target2...>] \n");
    printf("\n");
    printf("options:\n");
    printf("     -s <start>   - start image scan at block <start> (default is beginning)\n");
    printf("     -e <end>     - end image scan at block <start> (default is end)\n");
    printf("     -b blocksize - sets the blocksize (default is %d)\n",DEFAULT_BLOCKSIZE);
    printf("     -r           - prints the raw association map, in addition to the cleaned on\n");
    printf("     -S           - print stats\n");
    printf("     -MN	      - specifies size of bloom filter, in 2^N bits (default %d)\n",opt_M);
    printf("     -X1          - do not use bloom filter\n");
    
    exit(1);
}

/* A targetfile is like a blockfile, but it also keeps track of:
 * -  a sha1map, which is the block numbers in the target where each sha1 exists.
 * - a filemap, an array that holds all of the block lists
 */
class targetfile:public blockfile {
public:;
    sha1map_t sha1map;
    filemap_t filemap;
};

class targets_t:public std::vector<targetfile *>{
};

int main(int argc,char **argv)
{
    int opt_raw = 0;
    uint64_t opt_start = 0;
    uint64_t opt_end   = 0;
    int opt_stats = 0;
    int ch;
    bool use_bloom = 1;
    bool use_prefilter = 1;
    uint64_t bloom_false_positives=0;
    while ((ch = getopt(argc,argv,"s:e:rb:Sh?X:M:")) != -1){
	switch(ch){
	case 's': opt_start = atoi64(optarg);break;
	case 'e': opt_end   = atoi64(optarg);break;
	case 'r': opt_raw   = 1;break;
	case 'b': blocksize = atoi(optarg); break;
	case 'S': opt_stats++;break;
	case 'M': opt_M     = atoi(optarg); break;
	case 'X':
	    switch(atoi(optarg)){
	    case 1: use_bloom = 0;break;
	    case 2: use_prefilter = 0;break;
		break;
	    }
	    break;
	case 'h':
	case '?':
	default:
	    usage();
	}
    }
    if(opt_M <5 || opt_M > 32){
	fprintf(stderr,"M must be between 5 and 32.\n");
	exit(1);
    }

    argc -= optind; 
    argv += optind;
    
    if(argc<2) usage();

    if(blocksize<4){
	printf("Blocksizes smaller than 4 are meaningless. Try again.\n");
	exit(1);
    }

    printf("Creating bloom filter...\n");
    NSRLBloom b;
    if(b.create(0,160,opt_M,5,"In-memory bloom filter for frag find")){
	printf(">>> not enough memory for bloom filter\n");
	exit(1);
    }
    else {
	printf("Bloom filter created.\n");
    }


#ifdef SIGINFO
    signal(SIGINFO,status);
#endif
#ifdef USE_BITSET
    prefilter_t prefilter;	// bitset to hold first 3 bytes of block
#endif    
    blockfile imagefile;
    targets_t targets;

    if(imagefile.open(*argv++,blocksize)<0) err(1,"Cannot open %s",imagefile.filename);

    if(imagefile.filesize < blocksize){
	fprintf(stderr,"Image size (%"PRId64") is smaller than blocksize (%d); cannot continue.\n",
		targets[0]->filesize,blocksize);
	exit(1);
    }

    /* Load up all of the target files */
    while(*argv){
	targetfile &t = *(new targetfile());
	
	printf("\n");
	printf("Search file #%d\n",(int)targets.size()+1);
	printf("Filename: %s\n",*argv);

	if(t.open(*argv++,blocksize)<0){
	    err(1,"Error: Cannot open %s",targets[0]->filename);
	}
	
	if(imagefile.filesize < t.filesize ){
	    fprintf(stderr,"Error: Target file %s is larger than image file; cannot continue\n",t.filename);
	    exit(1);
	}
	    
	if(imagefile.filesize==0){
	    fprintf(stderr,"Error: Image size %s is zero?\n",t.filename);
	    exit(1);
	}

	printf("Blocks: %"PRId64" %d-byte blocks)\n", t.blocks,blocksize);

	/** Read through the target file block by block and create a SHA1 structure
	 * for each block that is read. Each SHA1 must be allocated because we need a place
	 * to keep all of these SHA1s. The map points to each SHA1.
	 */
	for(blocknumber=0;blocknumber < t.blocks; blocknumber++){
	    const u_char *buf = t.getblock(blocknumber);
	    if(use_prefilter) prefilter.setbuf(buf);
	
	    sha1_t *sha1 = new sha1_t(buf,blocksize);

	    // report if it is in the input file twice
	    if(opt_raw && b.query(sha1->sha1)){
		printf("Input file block %"PRId64" (%s) appears previously at ",blocknumber,sha1->hex());
		for(blocklist_t::const_iterator i = t.sha1map[*sha1].begin();
		    i!=t.sha1map[*sha1].end();
		    i++){
		    printf("%"PRId64" ",*i);
		}
		printf("\n");
	    }

	    // add to the map and to the bloom filter
	    t.sha1map[*sha1].push_back(blocknumber);
	    b.add(sha1->sha1);
	}
	t.filemap.resize(t.blocks);
	targets.push_back(&t);
    }

    printf("Now searching image file...\n");

    uint64_t count = 0;

    /** Read through the image file block by block. This time we can use a shared SHA1 structure
     * because we do not need to allocate a hash for each block.
     */

    if(opt_end==0) opt_end = imagefile.blocks;
    aftimer timer;
    timer.start();
    for(uint64_t blocknumber=opt_start;blocknumber < opt_end && blocknumber < imagefile.blocks; blocknumber++){
	if(blocknumber>opt_start && (((blocknumber-opt_start) % 100000)==0)){
	    int64_t blocks = blocknumber-opt_start;
	    int64_t total = opt_end-opt_start;
	    printf("%"PRId64"M out of %"PRId64"M sectors processed; hits=%"PRId64,
		   blocks/1000000,total/1000000,count);
	    if(timer.elapsed_seconds()>1.0){

		int tseconds = timer.eta((double)blocks/(double)total);
		int tminutes =  tseconds / 60;
		int days = tseconds / (24*60*60);
		int hours = (tminutes / 60) % 24;
		int minutes = tminutes % 60;
		int seconds = tseconds % 60;

		printf("; %3g Kblocks/sec; done in ", (blocks/1000)/timer.elapsed_seconds());
		if(days) printf("%d days ",days);
		printf("%d:%02d:%02d ",hours,minutes,seconds);
	    }
	    putchar('\n');
	    fflush(stdout);
	}

	/* Scan through the input file and the SHA1 of each block */
	const u_char *buf = imagefile.getblock(blocknumber);

#ifdef USE_BITSET
	/* If block is not in prefilter, then there is no sense to compute hash */
	if(use_prefilter){
	    if(prefilter.querybuf(buf)==0){
		continue;
	    }
	}
#endif
	/* Compute the SHA1 */
	struct sha1_t sha1(buf,blocksize);

	/* If the block's SHA1 is not in the bloom filter, there is no need to scan for it in the file maps. */

	if(b.query(sha1.sha1)==0) continue;

	/* Get the list of blocks where that SHA1 was observed and each one to the filemap.
	 */
	if(opt_raw && use_bloom) printf("Block %"PRId64" SHA1 in bloom filter.\n",blocknumber);
	if(opt_raw) printf("   >>> found at ");
	int64_t added = 0;
	for(targets_t::iterator t = targets.begin(); t!=targets.end(); t++){
	    for(blocklist_t::const_iterator i = (*t)->sha1map[sha1].begin();
		i!=(*t)->sha1map[sha1].end();
		i++){
		if(opt_raw) printf("   %"PRId64" ",*i);
		(*t)->filemap[*i].push_back(blocknumber);
		added++;
	    }
	}
	if(opt_raw) printf("\n");
	if(added==0) bloom_false_positives++;
	count++;
    }

    printf("\n\n");
    printf("Image file:  %s  (%"PRId64" blocks)\n",imagefile.filename,imagefile.blocks);
    printf("Blocksize: %d\n",targets[0]->blocksize);

    for(targets_t::iterator t = targets.begin(); t!=targets.end(); t++){
	printf("-------------------\n");
	printf("Target file: %s  (%"PRId64" blocks)\n",(*t)->filename,(*t)->blocks);
	if(opt_start!=0){
	    printf("   NOTE: Image scan started at block %"PRId64"\n",opt_start);
	}
	if(opt_end!=imagefile.blocks){
	    printf("   NOTE: Image scan ended at block %"PRId64"\n",opt_end);
	}

	printf("Blocks of target file found in image file: %"PRId64"\n",count);
	printf("Here is where they were found:\n");

	if(opt_raw){
	    printf("RAW FILEMAP:\n");
	    (*t)->filemap.print();
	    printf("===============================================================================\n");
	    printf("\n");
	}

	(*t)->filemap.clean();

	/* note: perhaps it would make more sense to iteratively clean. */
	(*t)->filemap.print();
	printf("Total blocks of original file found: %"PRId64" (%2.0f%%)\n",
	       (*t)->filemap.found_count(),
	       (*t)->filemap.found_percent());
	if(opt_stats){
	    double runtime = timer.elapsed_seconds();
	    printf("Total number of prefilter hits: %"PRId64" (%"PRId64" blocks filtered out)\n",
		   prefilter.hits,(imagefile.blocks-opt_start)-prefilter.hits);
	    printf("Total number of bloom filter hits: %"PRId64" (%"PRId64" blocks filtered out)\n",
		   b.hits,(imagefile.blocks-opt_start)-b.hits);
	    printf("Total number of bloom false positives: %"PRId64" (should be low)\n",bloom_false_positives);
	    printf("Total number of SHA1 codes calculated: %"PRId64" (should be low)\n",sha1_ctr);
	    printf("Prefilter Utilization: %5.2f%% (should be high)\n",prefilter.utilization()*100.0);
	    printf("Bloom filter utilization: %5.2f%% (ideally less than 50%%)\n",b.utilization()*100.0);
	    printf("Time to run: %4.2g seconds\n",runtime);
	}
    }
    return 0;
}
